use std::time::Duration;

use bevy::prelude::*;
use bevy_inspector_egui::{Inspectable, RegisterInspectable};
use bevy_prototype_lyon::{entity::ShapeBundle, prelude::*, shapes::Line};
use bevy_tweening::{
    component_animator_system, lens::TransformPositionLens, Animator, EaseFunction, Lens, Tween,
    TweenCompleted, TweeningPlugin, TweeningType,
};

use crate::turtle_shapes;

pub struct TurtlePlugin;

impl Plugin for TurtlePlugin {
    fn build(&self, app: &mut bevy::prelude::App) {
        app.add_plugin(TweeningPlugin)
            .add_startup_system(setup)
            .add_system(keypresses)
            .add_system(draw_lines)
            .add_system(component_animator_system::<Path>)
            .register_inspectable::<Colors>()
            .register_inspectable::<TurtleCommands>();
    }
}
#[derive(Bundle)]
pub struct Turtle {
    colors: Colors,
    commands: TurtleCommands,
    name: Name,
}

impl Default for Turtle {
    fn default() -> Self {
        Self {
            colors: Colors {
                color: Color::DARK_GRAY,
                fill_color: Color::BLACK,
            },
            commands: TurtleCommands::new(vec![
                TurtleCommand::Forward(Length(100.)),
                TurtleCommand::Right(Angle(90.)),
                TurtleCommand::Backward(Length(100.)),
                TurtleCommand::Right(Angle(90.)),
                TurtleCommand::Forward(Length(100.)),
                TurtleCommand::Right(Angle(45.)),
                TurtleCommand::Forward(Length(100.)),
                TurtleCommand::Right(Angle(90.)),
                TurtleCommand::Forward(Length(50.)),
                TurtleCommand::Right(Angle(90.)),
                TurtleCommand::Forward(Length(100.)),
                TurtleCommand::Right(Angle(90.)),
                TurtleCommand::Forward(Length(50.)),
                TurtleCommand::Right(Angle(45.)),
                TurtleCommand::Forward(Length(100.)),
                TurtleCommand::Left(Angle(120.)),
                TurtleCommand::Forward(Length(100.)),
                TurtleCommand::Left(Angle(120.)),
                TurtleCommand::Forward(Length(100.)),
                TurtleCommand::Right(Angle(150.)),
                TurtleCommand::Forward(Length(100.)),
            ]),
            name: Name::new("Turtle"),
        }
    }
}

impl Turtle {
    /*     pub fn set_color(&mut self, color: Color) {
        self.colors.color = color;
    }
    pub fn set_fill_color(&mut self, color: Color) {
        self.colors.fill_color = color;
    }
    pub fn get_colors(&self) -> &Colors {
        &self.colors
    }
    pub fn forward(&mut self) -> &mut Self {
        self.commands
            .commands
            .push(TurtleCommand::Forward(Length(100.0)));
        self
    } */
}

#[derive(Component, Inspectable)]
pub struct TurtleCommands {
    commands: Vec<TurtleCommand>,
    lines: Vec<TurtleGraphElement>,
    state: TurtleState,
}
#[derive(Inspectable)]
pub struct TurtleState {
    pub start: Vec2,
    pub heading: f32,
    pub index: u64,
}

impl TurtleCommands {
    fn new(commands: Vec<TurtleCommand>) -> Self {
        Self {
            commands,
            lines: vec![],
            state: TurtleState {
                start: Vec2::ZERO,
                heading: 0f32.to_radians(),
                index: 0,
            },
        }
    }
}

impl TurtleCommands {
    fn get_next(&mut self) -> Option<(Tween<Transform>, TurtleGraphElement)> {
        let index = self.state.index;
        let next_index = index + 1;

        if let Some(command) = self.commands.get(self.state.index as usize) {
            let res = match command {
                TurtleCommand::Forward(Length(x)) => {
                    crate::turtle_movement::turtle_move(&mut self.state, *x as f32)
                }
                TurtleCommand::Backward(Length(x)) => {
                    crate::turtle_movement::turtle_move(&mut self.state, -*x as f32)
                }
                TurtleCommand::Left(Angle(x)) => {
                    crate::turtle_movement::turtle_turn(&mut self.state, *x as f32)
                }
                TurtleCommand::Right(Angle(x)) => {
                    crate::turtle_movement::turtle_turn(&mut self.state, -*x as f32)
                }
                TurtleCommand::PenUp => todo!(),
                TurtleCommand::PenDown => todo!(),
                TurtleCommand::Circle => todo!(),
                TurtleCommand::Pause => todo!(),
            };
            self.state.index = next_index;
            Some(res)
        } else {
            None
        }
    }
}

#[derive(Inspectable, Default)]
pub enum TurtleGraphElement {
    TurtleLine {
        start: Vec2,
        end: Vec2,
    },
    #[default]
    Noop,
}

#[derive(Clone, Component, Inspectable)]
pub struct TurtleShape;

#[derive(Clone, Component, Inspectable)]
pub struct Colors {
    color: Color,
    fill_color: Color,
}

#[derive(Inspectable, Default)]
pub struct Length(f64);
#[derive(Inspectable, Default)]
pub struct Angle(f64);

#[derive(Component, Inspectable, Default)]
pub enum TurtleCommand {
    Forward(Length),
    Backward(Length),
    Left(Angle),
    Right(Angle),
    PenUp,
    PenDown,
    #[default]
    Pause,
    Circle,
}

fn setup(mut commands: Commands) {
    let animator = Animator::new(Tween::new(
        // Use a quadratic easing on both endpoints.
        EaseFunction::QuadraticInOut,
        // Loop animation back and forth.
        TweeningType::PingPong,
        // Animation time (one way only; for ping-pong it takes 2 seconds
        // to come back to start).
        Duration::from_millis(500),
        // The lens gives access to the Transform component of the Entity,
        // for the Animator to animate it. It also contains the start and
        // end values respectively associated with the progress ratios 0. and 1.
        TransformPositionLens {
            start: Vec3::ZERO,
            end: Vec3::new(40., 40., 0.),
        },
    ));
    commands.spawn_bundle(Camera2dBundle::default());
    commands
        .spawn_bundle(Turtle::default())
        .insert_bundle(GeometryBuilder::build_as(
            &turtle_shapes::turtle(),
            DrawMode::Outlined {
                fill_mode: FillMode::color(Color::MIDNIGHT_BLUE),
                outline_mode: StrokeMode::new(Color::BLACK, 1.0),
            },
            Transform::identity(),
        ))
        .insert(animator)
        .insert(TurtleShape);
}

struct MyCustomLens {
    start: Vec2,
    end: Vec2,
}

impl Lens<Path> for MyCustomLens {
    fn lerp(&mut self, target: &mut Path, ratio: f32) {
        let line = shapes::Line(self.start, self.start + ((self.end - self.start) * ratio));
        *target = ShapePath::build_as(&line);
    }
}

#[derive(Bundle)]
struct TurtleDrawLine {
    #[bundle]
    line: ShapeBundle,
    name: Name,
    marker: LineMarker,
}

#[derive(Component, Default)]
struct LineMarker;

impl TurtleDrawLine {
    fn new(start: Vec2, _end: Vec2, index: u64) -> Self {
        Self {
            line: GeometryBuilder::build_as(
                &Line(start, start),
                DrawMode::Outlined {
                    fill_mode: FillMode::color(Color::MIDNIGHT_BLUE),
                    outline_mode: StrokeMode::new(Color::BLACK, 1.0),
                },
                Transform::identity(),
            ),
            name: Name::new(format!("Line {}", index)),
            marker: LineMarker,
        }
    }
}

fn draw_lines(
    mut commands: Commands,
    mut tcmd: Query<&mut TurtleCommands>,
    mut turtle: Query<&mut Animator<Transform>, With<TurtleShape>>,
    mut query_event: EventReader<TweenCompleted>, // TODO: howto attach only to the right event?
) {
    for _ev in query_event.iter() {
        let mut tcmd = tcmd.single_mut();
        if let Some((turtle_animation, graph_element_to_draw)) = tcmd.get_next() {
            let mut turtle = turtle.single_mut();
            turtle.set_tweenable(turtle_animation);
            match graph_element_to_draw {
                TurtleGraphElement::TurtleLine { start, end } => {
                    let line_animator = Animator::new(Tween::new(
                        // Use a quadratic easing on both endpoints.
                        EaseFunction::QuadraticInOut,
                        // Loop animation back and forth.
                        TweeningType::Once,
                        // Animation time (one way only; for ping-pong it takes 2 seconds
                        // to come back to start).
                        Duration::from_millis(500),
                        // The lens gives access to the Transform component of the Entity,
                        // for the Animator to animate it. It also contains the start and
                        // end values respectively associated with the progress ratios 0. and 1.
                        MyCustomLens { start, end },
                    ));
                    commands
                        .spawn_bundle(TurtleDrawLine::new(start, end, tcmd.state.index))
                        .insert(line_animator);
                }
                TurtleGraphElement::Noop => println!("No drawing!"),
            }
        } else {
            println!("nothing to draw")
        };
    }
}

fn keypresses(
    mut commands: Commands,
    keys: Res<Input<KeyCode>>,
    mut qry: Query<&mut Animator<Transform>, With<TurtleShape>>,
    mut tcmd: Query<&mut TurtleCommands>,
) {
    if keys.just_pressed(KeyCode::W) {
        let mut tcmd = tcmd.single_mut();
        tcmd.state = TurtleState {
            start: Vec2::ZERO,
            heading: 0.,
            index: 0,
        };
        if let Some((turtle_animation, graph_element_to_draw)) = tcmd.get_next() {
            let mut shap = qry.single_mut();
            shap.set_tweenable(turtle_animation);
            match graph_element_to_draw {
                TurtleGraphElement::TurtleLine { start, end } => {
                    let line_animator = Animator::new(Tween::new(
                        // Use a quadratic easing on both endpoints.
                        EaseFunction::QuadraticInOut,
                        // Loop animation back and forth.
                        TweeningType::Once,
                        // Animation time (one way only; for ping-pong it takes 2 seconds
                        // to come back to start).
                        Duration::from_millis(500),
                        // The lens gives access to the Transform component of the Entity,
                        // for the Animator to animate it. It also contains the start and
                        // end values respectively associated with the progress ratios 0. and 1.
                        MyCustomLens { start, end },
                    ));
                    commands
                        .spawn_bundle(TurtleDrawLine::new(start, end, tcmd.state.index))
                        .insert(line_animator);
                }
                TurtleGraphElement::Noop => (),
            }
        };
    }
}
