use std::{f32::consts::PI, time::Duration};

use bevy::prelude::{Transform, Vec2};
use bevy_tweening::{
    lens::{TransformPositionLens, TransformRotateZLens},
    EaseFunction, Tween, TweeningType,
};

use crate::turtle::{TurtleGraphElement, TurtleState};

pub fn turtle_turn(
    state: &mut TurtleState,
    angle_to_turn: f32,
) -> (Tween<Transform>, TurtleGraphElement) {
    let start = state.heading;
    let end = state.heading + (angle_to_turn * PI / 180.);
    let animation = Tween::new(
        // Use a quadratic easing on both endpoints.
        EaseFunction::QuadraticInOut,
        // Loop animation back and forth.
        TweeningType::Once,
        // Animation time (one way only; for ping-pong it takes 2 seconds
        // to come back to start).
        Duration::from_millis(500),
        // The lens gives access to the Transform component of the Entity,
        // for the Animator to animate it. It also contains the start and
        // end values respectively associated with the progress ratios 0. and 1.
        TransformRotateZLens { start, end },
    )
    .with_completed_event(state.index as u64);
    let line = TurtleGraphElement::Noop;
    state.heading = end % (2. * PI);
    (animation, line)
}

pub fn turtle_move(state: &mut TurtleState, length: f32) -> (Tween<Transform>, TurtleGraphElement) {
    let start = state.start;
    let end = state.start + (Vec2::from_angle(state.heading) * length);
    let turtle_movement_animation = Tween::new(
        // accelerate and decelerate
        EaseFunction::QuadraticInOut,
        // Loop animation back and forth.
        TweeningType::Once,
        // later to be controlled by speed
        Duration::from_millis(500),
        // set the start and end of the animation
        TransformPositionLens {
            start: start.extend(0.),
            end: end.extend(0.),
        },
    )
    .with_completed_event(state.index as u64);
    let line = TurtleGraphElement::TurtleLine { start, end };
    state.start = end;
    (turtle_movement_animation, line)
}
